/* Copyright 1997-2008 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#include <alloca.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <glob.h>
#include <libintl.h>
#include <locale.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* Changes
   1998-09-22 - Arnaldo Carvalho de Melo <acme@conectiva.com.br>
                i18n for init.d scripts (eg.: description(pt_BR) is a brazilian
		portuguese description for the package)
*/

#define _(String) gettext((String))

#include "leveldb.h"
#include "xmalloc.h"

int skipDirEntry(const struct dirent *ent)
{
    const char *dn;

    /* Skip any file starting with a . */
    if (ent->d_name[0] == '.')
	return 1;

    /* Skip files with known bad extensions */
    if ((dn = strrchr(ent->d_name, '.')) != NULL &&
	(!strcmp(dn, ".rpmsave") || !strcmp(dn, ".rpmnew") ||
	 !strcmp(dn, ".rpmorig") || !strcmp(dn, ".swp")))
	return 1;

    dn = ent->d_name + strlen(ent->d_name) - 1;
    if (*dn == '~' || *dn == ',')
	return 1;

    return 0;
}

int parseLevels(const char *str, int emptyOk) {
    const char *chptr;
    int rc = 0;

    if (!str || !*str)
	return emptyOk ? 0 : -1;

    for (chptr = str; *chptr; ++chptr) {
	if (!isdigit(*chptr) || *chptr > '6') return -1;
	rc |= 1 << (*chptr - '0');
    }

    return rc;
}

int readDescription(char *start, char *bufstop, char **english_desc, char **serv_desc) {
	int english;
	int is_my_lang = 0;
	char * lang = getenv ("LANG");
	char * end, *next;
	char *desc;

	english = *start == ':';
	end = strchr(start, '\n');
	if (!end)
	    next = end = bufstop;
	else
	    next = end + 1;

	if (!english) {
		char *final_parenthesis;

		if (*start != '(') {
		    return 1;
		}

                ++start;
		final_parenthesis = strchr (start, ')');

		if (final_parenthesis == NULL || final_parenthesis - start > 5) {
		    return 1;
		}

		is_my_lang = lang ? strncmp (lang, start, strlen (lang)) == 0 : 0;
		start = final_parenthesis + 1;
		if (*start == ':')
		    ++start;
	} else ++start;

	while (start < end && isspace(*start))
		start++;
	if (start >= end)
		return 1;

	desc = xmalloc(end - start + 1);
	strncpy(desc, start, end - start);
	desc[end - start] = '\0';

	for (start = next; desc[strlen(desc) - 1] == '\\'; start = next) {
		size_t i;

		desc[strlen(desc) - 1] = '\0';

		while (start < bufstop && isspace(*start)) start++;
		if (start >= bufstop || *start != '#')
		    return 1;

		start++;

		while (start < bufstop && isspace(*start)) start++;
		if (start >= bufstop)
		    return 1;

		end = strchr(start, '\n');
		if (!end)
		    next = end = bufstop;
		else
		    next = end + 1;

		i = strlen(desc);
		desc = xrealloc(desc, i + end - start + 1);
		strncat(desc, start, end - start);
	}

	if (desc) {
		if (is_my_lang) {
			free(*serv_desc);
			*serv_desc = desc;
		} else if (english) {
			*serv_desc = xfree(*serv_desc);
			free(*english_desc);
			*english_desc = desc;
		} else free(desc);
	}

	return 0;
}

static  ssize_t
read_loop(int fd, char *buffer, size_t count)
{
	ssize_t offset = 0;

	while (count > 0) {
		ssize_t block =
			TEMP_FAILURE_RETRY(read(fd, &buffer[offset], count));

		if (block <= 0)
			return offset ? offset : block;
		offset += block;
		count -= block;
	}
	return offset;
}

static  ssize_t
write_loop(int fd, const char *buffer, size_t count)
{
	ssize_t offset = 0;

	while (count > 0) {
		ssize_t block =
			TEMP_FAILURE_RETRY(write(fd, &buffer[offset], count));

		if (block <= 0)
			return offset ? : block;
		offset += block;
		count -= block;
	}
	return offset;
}

int readXinetdServiceInfo(char *name, struct service * service) {
	char * filename;
	int fd;
	struct service serv = {
			name: NULL,
			levels: -1,
			kPriority: 100,
			sPriority: -1,
			desc: NULL,
			startDeps: NULL,
			stopDeps: NULL,
			softStartDeps: NULL,
			softStopDeps: NULL,
		        provides: NULL,
			type: TYPE_XINETD,
			isLSB: 0,
			enabled: -1
	};
	struct stat sb;
	char * buf = NULL, *ptr;
	char * eng_desc = NULL, *end_buf;

	xasprintf(&filename, XINETDDIR "/%s", name);

	if ((fd = open(filename, O_RDONLY)) < 0) goto out_err;
	if (fstat(fd,&sb) || !S_ISREG(sb.st_mode)) goto out_err;
	if (!(buf = malloc(sb.st_size+1))) goto out_err;
	if (read_loop(fd,buf,sb.st_size) != sb.st_size) goto out_err;
	close(fd);
	fd = -1;
	serv.name = xstrdup(name);
	end_buf = buf + sb.st_size;
	*end_buf = '\0';
	for (; buf; buf = ptr) {
		while (*buf == '\n')
			buf++;
		ptr = strchr(buf, '\n');
		if (*buf == '#') {
			buf++;
			while (isspace(*buf) && *buf != '\n')
				buf++;
			if (!strncmp(buf, "default:", 8)) {
				buf += 8;
				while (isspace(*buf) && *buf != '\n')
					buf++;
				serv.enabled = !strncmp(buf, "on", 2);
			} else if (!strncmp(buf, "description:", 12)) {
				buf += 11;
				if (readDescription(buf, end_buf,
						    &eng_desc, &serv.desc)) {
					serv.desc = xfree(serv.desc);
				}
			}
			continue;
		}
		while (isspace(*buf) && *buf != '\n')
			buf++;
		if (!strncmp(buf, "disable", 7)) {
			buf += 7;
			while (isspace(*buf) && *buf != '\n')
				buf++;
			if (*buf == '=') {
				buf++;
				while (isspace(*buf) && *buf != '\n')
					buf++;
			} else
				buf = NULL;

			if (buf && strncmp(buf, "yes", 3)) {
				serv.levels = parseLevels("0123456",0);
				if (serv.enabled == -1)
				  serv.enabled = 1;
			} else {
				serv.levels = 0;
				if (serv.enabled == -1)
				  serv.enabled = 0;
			}
		}
	}

	if (!serv.desc) {
		if (eng_desc)
			serv.desc = eng_desc;
		else
			serv.desc = xstrdup(name);
	} else
		eng_desc = xfree(eng_desc);

	*service = serv;
	return 0;
out_err:
        if (fd >= 0)
            close(fd);
        free(buf);
        free(filename);
        return -1;
}

int readServices(struct service **services) {
	DIR * dir;
	struct dirent * ent;
	struct stat sb;
	struct service *servs = NULL;
	int numservs = 0;
	char fn[1024];

	if (!(dir = opendir(RUNLEVELS "/init.d"))) {
		fprintf(stderr, _("failed to open %s/init.d: %s\n"), RUNLEVELS,
			strerror(errno));
		return 1;
	}

	while ((ent = readdir(dir))) {
		if (skipDirEntry(ent))
		    continue;

		sprintf(fn, RUNLEVELS "/init.d/%s", ent->d_name);
		if (stat(fn, &sb)) {
			continue;
		}
		if (!S_ISREG(sb.st_mode) || !(sb.st_mode & S_IXUSR)) continue;
		servs = xrealloc(servs, (numservs+1) * sizeof(struct service));
		if (!readServiceInfo(ent->d_name, TYPE_INIT_D, servs + numservs, 0))
			numservs++;
	}
	*services = servs;
	return numservs;
}

int readServiceInfo(char * name, int type, struct service * service, int honorHide) {
    char * filename = NULL;
    int fd;
    struct service serv, serv_overrides;
    int parseret;

    if (!(type & TYPE_INIT_D))
	goto try_xinetd;

    xasprintf(&filename, RUNLEVELS "/init.d/%s", name);

    if ((fd = open(filename, O_RDONLY)) < 0)
	goto try_xinetd;

    free(filename);
    parseret = parseServiceInfo(fd, name, &serv, honorHide, 0);
    if (parseret)
        return parseret;

    xasprintf(&filename, RUNLEVELS "/chkconfig.d/%s", name);
    if ((fd = open(filename, O_RDONLY)) >= 0) {
        parseret = parseServiceInfo(fd, name, &serv_overrides, honorHide, 1);
        if (parseret >= 0) {
            if (serv_overrides.name) serv.name = serv_overrides.name;
            if (serv_overrides.levels != -1) serv.levels = serv_overrides.levels;
            if (serv_overrides.kPriority != -2) serv.kPriority = serv_overrides.kPriority;
            if (serv_overrides.sPriority != -2) serv.sPriority = serv_overrides.sPriority;
            if (serv_overrides.desc) serv.desc = serv_overrides.desc;
            if (serv_overrides.startDeps) serv.startDeps = serv_overrides.startDeps;
            if (serv_overrides.stopDeps) serv.stopDeps = serv_overrides.stopDeps;
            if (serv_overrides.softStartDeps) serv.softStartDeps = serv_overrides.softStartDeps;
            if (serv_overrides.softStopDeps) serv.softStopDeps = serv_overrides.softStopDeps;
            if (serv_overrides.provides) serv.provides = serv_overrides.provides;
            if (serv_overrides.isLSB || serv.isLSB) serv.isLSB = 1;
        }
    }
    serv.currentLevels = whatLevels(name);

    free(filename);
    *service = serv;
    return 0;

try_xinetd:
    free(filename);
    if (!(type & TYPE_XINETD))
	return -1;
    return readXinetdServiceInfo(name,service);
}

int readServiceDifferences(char * name, int type, struct service * service, struct service * service_overrides, int honorHide) {
    char * filename = NULL;
    int fd;
    struct service serv, serv_overrides;
    int parseret;

    if (!(type & TYPE_INIT_D))
	goto try_xinetd;

    xasprintf(&filename, RUNLEVELS "/init.d/%s", name);

    if ((fd = open(filename, O_RDONLY)) < 0) {
	goto try_xinetd;
    }

    free(filename);
    parseret = parseServiceInfo(fd, name, &serv, honorHide, 0);
    if (parseret) {
        return parseret;
    }

    xasprintf(&filename, RUNLEVELS "/chkconfig.d/%s", name);
    if ((fd = open(filename, O_RDONLY)) >= 0) {
        parseret = parseServiceInfo(fd, name, &serv_overrides, honorHide, 1);
    } else {
        parseret = 1;
    }

    free(filename);
    if (parseret) {
        return 1;
    }

    *service = serv;
    *service_overrides = serv_overrides;
    return 0;

try_xinetd:
    free(filename);
    if (!(type & TYPE_XINETD))
	return -1;
    return readXinetdServiceInfo(name,service);
}

static struct dep *parseDeps(char *pos, char *end) {
    char *t;
    int numdeps = 0;
    struct dep *deps = NULL;

    while (1) {
	while (*pos && isspace(*pos) && pos < end) pos++;
	if (pos == end)
            break;
        t = pos;
        while (*t && !isspace(*t) && t < end) t++;
        if (isspace(*t)) {
            *t = '\0';
            t++;
        }
        numdeps++;
        deps = xrealloc(deps,
		 (numdeps + 1) * sizeof(struct dep));
        deps[numdeps-1].name = xstrdup(pos);
        deps[numdeps-1].handled = 0;
        memset(&deps[numdeps],'\0',sizeof(struct dep));
        if (!t || t >= end)
            break;
        else
            pos = t;
    }
    return deps;
}


int parseServiceInfo(int fd, char * name, struct service * service, int honorHide, int partialOk) {
    struct stat sb;
    char * bufstart, * bufstop, * start, * end, * next;
    struct service serv = {
		    name: NULL,
		    levels: -1,
		    kPriority: 100,
		    sPriority: -1,
		    currentLevels: 0,
		    desc: NULL,
		    startDeps: NULL,
		    stopDeps: NULL,
		    softStartDeps: NULL,
		    softStopDeps: NULL,
		    provides: NULL,
		    type: TYPE_INIT_D,
		    isLSB: 0,
		    enabled: 0
    };
    char * english_desc = NULL;

    if (fstat(fd, &sb) || !S_ISREG(sb.st_mode) || !(sb.st_mode & S_IXUSR)) {
	close(fd);
	return -1;
    }

    if (!(bufstart = malloc(sb.st_size + 1))) {
	close(fd);
	return -1;
    }

    if (read_loop(fd, bufstart, sb.st_size) != sb.st_size) {
	free(bufstart);
	close(fd);
	return -1;
    }

    close(fd);

    bufstop = bufstart + sb.st_size;
    *bufstop = 0;

    for (start = bufstart;
	 start < bufstop && (serv.levels == -1 || !serv.desc);
	 start = next) {
	while (isspace(*start) && start < bufstop) start++;
	if (start == bufstop) break;

	end = strchr(start, '\n');
	if (!end)
	    next = end = bufstop;
	else
	    next = end + 1;

	if (*start != '#') continue;

	start++;
	if (!strncmp(start, "## BEGIN INIT INFO", 18)) {
	    serv.isLSB = 1;
	    continue;
	}

	while (isspace(*start) && start < end) start++;
	if (start == end) continue;
	if (honorHide && !strncmp(start, "hide:", 5)) {
	    start += 5;
	    while (isspace(*start) && start < end) start++;
	    if (start == end || !strncmp(start, "true", 4)) {
		serv.desc = xfree(serv.desc);
		free(bufstart);
		return 1;
	    }
	}

	if (!strncmp(start, "chkconfig:", 10)) {
	    int spri, kpri;
	    char levelbuf[16];
	    char overflow;

	    start += 10;
	    while (isspace(*start) && start < end) start++;
	    if (start == end) {
		serv.desc = xfree(serv.desc);
		free(bufstart);
		return 1;
	    }

	    if ((sscanf(start, "%15s %d %d%c", levelbuf,
			&spri, &kpri, &overflow) != 4) ||
		 !isspace(overflow)) {
		serv.desc = xfree(serv.desc);
		free(bufstart);
		return 1;
	    }
	    if (spri > 99 || kpri > 99 || kpri < 0 || spri < 0) {
		    serv.desc = xfree(serv.desc);
		    free(bufstart);
		    return 1;
	    }
	    if (serv.sPriority == -1)
			serv.sPriority = spri;
	    if (serv.kPriority == 100)
			serv.kPriority = kpri;
	    if (serv.levels == -1) {
		    if (!strcmp(levelbuf, "-"))
			    serv.levels = 0;
		    else
			    serv.levels = parseLevels(levelbuf, 0);
	    }
	    if (serv.levels == -1) {
		serv.desc = xfree(serv.desc);
		free(bufstart);
		return 1;
	    }
	} else if (!strncmp(start, "description", 11) ||
		   !strncmp(start, "Description:", 12)) {
		if (readDescription(start+11, bufstop, &english_desc, &serv.desc)) {
			serv.desc = xfree(serv.desc);
		}
	} else if (!strncmp(start, "Short-Description:", 18)) {
		if (readDescription(start+17, bufstop, &english_desc, &serv.desc)) {
			serv.desc = xfree(serv.desc);
		}
	} else if (!strncmp(start, "Default-Start:", 14)) {
		char *t;

		start+=14;
		while (1) {
			int lev;

			lev = strtol(start, &t, 10);
			if (t && t != start)
				start = t;
			else
				break;
			if (serv.levels == -1)
				serv.levels = 0;
			serv.levels |= 1 << lev;
		}
	} else if (!strncmp(start, "Default-Stop:", 13)) {
		char *t;

		start+=13;
		while (1) {
			int lev;

			lev = strtol(start, &t, 10);
			if (t && t != start)
				start = t;
			else
				break;
			if (serv.levels == -1)
				serv.levels = 0;
			serv.levels &= ~(1 << lev);
		}
	} else if (!strncmp(start, "Required-Start:", 15)) {
		start+=15;
		serv.startDeps = parseDeps(start, end);
	} else if (!strncmp(start, "Required-Stop:", 14)) {
		start+=14;
		serv.stopDeps = parseDeps(start, end);
	} else if (!strncmp(start, "Should-Start:", 13)) {
		start+=13;
		serv.softStartDeps = parseDeps(start, end);
	} else if (!strncmp(start, "Should-Stop:", 12)) {
		start+=12;
		serv.softStopDeps = parseDeps(start, end);
	} else if (!strncmp(start, "Provides:", 9)) {
		char *t;
		int numdeps = 0;

		start+=9;
		while (1) {
			while (*start && isspace(*start) && start < end) start++;
			if (start == end)
				break;
			t = start;
			while (*t && !isspace(*t) && t < end) t++;
			if (isspace(*t)) {
				*t = '\0';
				t++;
			}
			numdeps++;
			serv.provides = xrealloc(serv.provides,
						 (numdeps + 1) * sizeof(char *));
			serv.provides[numdeps-1] = xstrdup(start);
			serv.provides[numdeps] = NULL;
			if (!t || t >= end)
				break;
			else
				start = t;
		}

	}
    }

    free(bufstart);

    if (!serv.desc) {
      if (english_desc)
	serv.desc = english_desc;
      else
        serv.desc = xstrdup(name);
    } else
	free (english_desc);

    if (!partialOk && ((serv.levels == -1) || !serv.desc || (!serv.isLSB && (serv.sPriority == -1 || serv.kPriority == 100)))) {
	return 1;
    }

    serv.name = xstrdup(name);
    if (!serv.provides) {
	    serv.provides = xmalloc(2 * sizeof(char *));
	    serv.provides[0] = xstrdup(name);
	    serv.provides[1] = NULL;
    }

    *service = serv;
    return 0;
}

/* returns -1 on error */
int currentRunlevel(void) {
    FILE * p;
    char response[50];

    p = popen("/sbin/runlevel", "r");
    if (!p) return -1;

    if (!fgets(response, sizeof(response), p)) {
	pclose(p);
	return -1;
    }

    pclose(p);

    if (!response[0] || response[1] != ' ' ||
	!isdigit(response[2]) || response[3] != '\n')
	return -1;

    return response[2] - '0';
}

int findServiceEntries(const char *name, int level, glob_t * globresptr) {
    char *match;
    glob_t globres;
    int rc;

    xasprintf(&match, "%s/rc%d.d/[SK][0-9][0-9]%s", RUNLEVELS, level, name);

    rc = glob(match, GLOB_ERR | GLOB_NOSORT, NULL, &globres);

    if (rc && rc != GLOB_NOMATCH) {
	fprintf(stderr, _("failed to glob pattern %s: %s\n"), match,
		strerror(errno));
	free(match);
	return 1;
    } else if (rc == GLOB_NOMATCH) {
	globresptr->gl_pathc = 0;
	free(match);
	return 0;
    }

    *globresptr = globres;
    free(match);
    return 0;
}

int cleanServiceEntry(const char *name, int level) {
    glob_t globres;

    if (findServiceEntries(name, level, &globres))
	return 1;

    if (globres.gl_pathc) {
	size_t i;

	for (i = 0; i < globres.gl_pathc; ++i)
	    unlink(globres.gl_pathv[i]);
	globfree(&globres);
    }

    return 0;
}

int isConfigured(char * name, int level, int *priority, char *type) {
    glob_t globres;
    char *pri_string;

    if (findServiceEntries(name, level, &globres))
	exit(1);

    if (!globres.gl_pathc)
	return 0;

    if (type) {
        *type = globres.gl_pathv[0][11];
    }

    if (priority) {
        pri_string = strndup(globres.gl_pathv[0]+12, 2);
        if (!pri_string) return 0;
        sscanf(pri_string, "%d", priority);
        free(pri_string);
    }

    globfree(&globres);
    return 1;
}

int isOn(char * name, int level) {
    glob_t globres;
    int rc;

    if (level == -1) {
	level = currentRunlevel();
	if (level == -1) {
	    fprintf(stderr, _("cannot determine current run level\n"));
	    return 0;
	}
    }

    if (findServiceEntries(name, level, &globres))
	exit(1);

    if (!globres.gl_pathc)
	return 0;

    rc = !!strstr(globres.gl_pathv[0], "/S");
    globfree(&globres);
    return rc;
}

int whatLevels(char *name) {
    int i, ret = 0;

    for (i = 0; i < 7; i++) {
        ret |= (isOn(name, i) << i);
    }
    return ret;
}

int setXinetdService(struct service s, int on) {
	int fd = -1, rc = -1;
	char *oldfname = NULL, *newfname = NULL;
	char *bufstart = NULL, *bufstop, *buf, *ptr;
	struct stat sb;

	if (on == -1) {
		on = s.enabled ? 1 : 0;
	}

	xasprintf(&oldfname, "%s/%s", XINETDDIR, s.name);
	if ((fd = open(oldfname, O_RDONLY)) < 0)
		goto setXinetdService_done;
	if (fstat(fd, &sb) || !S_ISREG(sb.st_mode))
		goto setXinetdService_done;
	if (!(bufstart = malloc(sb.st_size+1)))
		goto setXinetdService_done;
	if (read_loop(fd, bufstart, sb.st_size) != sb.st_size)
		goto setXinetdService_done;
	close(fd), fd = -1;
	bufstop = bufstart + sb.st_size;
	*bufstop = '\0';

	xasprintf(&newfname, "%s.XXXXXX", oldfname);
	if ((fd = mkstemp(newfname)) < 0) {
		newfname = xfree(newfname);
		goto setXinetdService_done;
	}

	for (buf = bufstart; buf && buf < bufstop; buf = ptr) {
		char *tmp = buf;

		ptr = strchr(buf,'\n');
		if (ptr) {
			*ptr = '\0';
			ptr++;
		}
		while (isspace(*buf)) buf++;
		if (!buf[0] || !strncmp(buf, "disable", 7))
			continue;
		if (write_loop(fd, tmp, strlen(tmp)) != strlen(tmp) ||
		    write_loop(fd, "\n", 1) != 1)
			goto setXinetdService_done;
		if (buf[0] == '{') {
			const char *prefix = "\tdisable\t= ";
			const char *yesno = on ? "no\n" : "yes\n";
			if (write_loop(fd, prefix, strlen(prefix)) != strlen(prefix) ||
			    write_loop(fd, yesno, strlen(yesno)) != strlen(yesno))
				goto setXinetdService_done;
		}
	}
	fchmod(fd, sb.st_mode & 0666);
	close(fd), fd = -1;
	unlink(oldfname);
	rc = rename(newfname, oldfname);

setXinetdService_done:
	if (rc && newfname)
		unlink(newfname);
	if (fd >= 0) close(fd);
	free(newfname);
	free(bufstart);
	free(oldfname);
	return rc;
}

int doSetService(struct service s, int level, int on) {
    int priority = on ? s.sPriority : s.kPriority;
    int rc;
    char *linkname, *linkto;

    cleanServiceEntry(s.name, level);

    xasprintf(&linkname, "%s/rc%d.d/%c%02d%s", RUNLEVELS, level,
	      on ? 'S' : 'K', priority, s.name);
    xasprintf(&linkto, "../init.d/%s", s.name);

    unlink(linkname);	/* just in case */
    if ((rc = symlink(linkto, linkname))) {
	fprintf(stderr, _("failed to make symlink %s: %s\n"), linkname,
		strerror(errno));
    }

    free(linkto);
    free(linkname);
    return !!rc;
}

int systemdActive() {
    struct stat s;

    return !lstat("/run/systemd/system/", &s) &&
		!!S_ISDIR(s.st_mode);
}

int isOverriddenBySystemd(const char *service) {
    char *p;
    int rc;

    xasprintf(&p, SYSTEMD_SERVICE_PATH "/%s.service", service);
    rc = !access(p, F_OK);
    free(p);
    return rc;
}
